import { Line, registry, ScatterController, defaults, clipArea, unclipArea, requestAnimFrame, merge, LineController, Chart, LinearScale, Point } from '@sgratzl/chartjs-esm-facade';
import { forceSimulation, forceCenter, forceCollide, forceLink, forceManyBody, forceX, forceY, forceRadial } from 'd3-force';
import { hierarchy, tree, cluster } from 'd3-hierarchy';

function horizontal(from, to, options) {
  return {
    fx: (to.x - from.x) * options.tension,
    fy: 0,
    tx: (from.x - to.x) * options.tension,
    ty: 0,
  };
}

function vertical(from, to, options) {
  return {
    fx: 0,
    fy: (to.y - from.y) * options.tension,
    tx: 0,
    ty: (from.y - to.y) * options.tension,
  };
}

function radial(from, to, options) {
  const angleHelper = Math.hypot(to.x - from.x, to.y - from.y) * options.tension;
  return {
    fx: Number.isNaN(from.angle) ? 0 : Math.cos(from.angle) * angleHelper,
    fy: Number.isNaN(from.angle) ? 0 : Math.sin(from.angle) * -angleHelper,
    tx: Number.isNaN(to.angle) ? 0 : Math.cos(to.angle) * -angleHelper,
    ty: Number.isNaN(to.angle) ? 0 : Math.sin(to.angle) * angleHelper,
  };
}

class EdgeLine extends Line {
  draw(ctx) {
    const options = this.options;

    ctx.save();

    // Stroke Line Options
    ctx.lineCap = options.borderCapStyle;
    ctx.setLineDash(options.borderDash || []);
    ctx.lineDashOffset = options.borderDashOffset;
    ctx.lineJoin = options.borderJoinStyle;
    ctx.lineWidth = options.borderWidth;
    ctx.strokeStyle = options.borderColor;

    const orientations = {
      horizontal,
      vertical,
      radial,
    };
    const layout = orientations[this._orientation] || orientations.horizontal;

    const renderLine = (from, to) => {
      const shift = layout(from, to, options);

      const fromX = {
        cpx: from.x + shift.fx,
        cpy: from.y + shift.fy,
      };
      const toX = {
        cpx: to.x + shift.tx,
        cpy: to.y + shift.ty,
      };

      // Line to next point
      if (options.stepped === 'middle') {
        const midpoint = (from.x + to.x) / 2.0;
        ctx.lineTo(midpoint, from.y);
        ctx.lineTo(midpoint, to.y);
        ctx.lineTo(to.x, to.y);
      } else if (options.stepped === 'after') {
        ctx.lineTo(from.x, to.y);
        ctx.lineTo(to.x, to.y);
      } else if (options.stepped) {
        ctx.lineTo(to.x, from.y);
        ctx.lineTo(to.x, to.y);
      } else if (options.tension) {
        ctx.bezierCurveTo(fromX.cpx, fromX.cpy, toX.cpx, toX.cpy, to.x, to.y);
      } else {
        ctx.lineTo(to.x, to.y);
      }
      return to;
    };

    const source = this.source.getProps(['x', 'y', 'angle']);
    const target = this.target.getProps(['x', 'y', 'angle']);
    const points = this.getProps(['points']).points;

    // Stroke Line
    ctx.beginPath();

    let from = source;
    ctx.moveTo(from.x, from.y);
    if (points && points.length > 0) {
      from = points.reduce(renderLine, from);
    }
    renderLine(from, target);

    ctx.stroke();

    if (options.directed) {
      const to = target;
      // compute the rotation based on from and to
      const shift = layout(from, to, options);
      const s = options.arrowHeadSize;
      const offset = options.arrowHeadOffset;
      ctx.save();
      ctx.translate(to.x, target.y);
      if (options.stepped === 'middle') {
        const midpoint = (from.x + to.x) / 2.0;
        ctx.rotate(Math.atan2(to.y - to.y, to.x - midpoint));
      } else if (options.stepped === 'after') {
        ctx.rotate(Math.atan2(to.y - to.y, to.x - from.x));
      } else if (options.stepped) {
        ctx.rotate(Math.atan2(to.y - from.y, to.x - to.x));
      } else if (options.tension) {
        const toX = {
          x: to.x + shift.tx,
          y: to.y + shift.ty,
        };
        const f = 0.1;
        ctx.rotate(Math.atan2(to.y - toX.y * (1 - f) - from.y * f, to.x - toX.x * (1 - f) - from.x * f));
      } else {
        ctx.rotate(Math.atan2(to.y - from.y, to.x - from.x));
      }
      ctx.translate(-offset, 0);
      ctx.beginPath();

      ctx.moveTo(0, 0);
      ctx.lineTo(-s, -s / 2);
      ctx.lineTo(-s * 0.9, 0);
      ctx.lineTo(-s, s / 2);
      ctx.closePath();
      ctx.fillStyle = ctx.strokeStyle;
      ctx.fill();

      ctx.restore();
    }

    ctx.restore();

    // point helper
    // ctx.save();
    // ctx.strokeStyle = 'blue';
    // ctx.beginPath();
    // ctx.moveTo(from.x, from.y);
    // ctx.lineTo(from.x + shift.fx, from.y + shift.fy, 3, 3);
    // ctx.stroke();
    // ctx.strokeStyle = 'red';
    // ctx.beginPath();
    // ctx.moveTo(to.x, to.y);
    // ctx.lineTo(to.x + shift.tx, to.y + shift.ty, 3, 3);
    // ctx.stroke();
    // ctx.restore();
  }
}

EdgeLine.id = 'edgeLine';
EdgeLine.defaults = /*#__PURE__*/ Object.assign({}, Line.defaults, {
  tension: 0,
  directed: false,
  arrowHeadSize: 15,
  arrowHeadOffset: 5,
});
EdgeLine.defaultRoutes = Line.defaultRoutes;

/*!
 * Chart.js v3.0.0-alpha.2
 * https://www.chartjs.org
 * (c) 2020 Chart.js Contributors
 * Released under the MIT License
 */
function _capitalize(str) {
	return str.charAt(0).toUpperCase() + str.slice(1);
}

/*!
 * Chart.js v3.0.0-alpha.2
 * https://www.chartjs.org
 * (c) 2020 Chart.js Contributors
 * Released under the MIT License
 */
const arrayEvents = ['push', 'pop', 'shift', 'splice', 'unshift'];
function listenArrayEvents(array, listener) {
	if (array._chartjs) {
		array._chartjs.listeners.push(listener);
		return;
	}
	Object.defineProperty(array, '_chartjs', {
		configurable: true,
		enumerable: false,
		value: {
			listeners: [listener]
		}
	});
	arrayEvents.forEach((key) => {
		const method = '_onData' + _capitalize(key);
		const base = array[key];
		Object.defineProperty(array, key, {
			configurable: true,
			enumerable: false,
			value(...args) {
				const res = base.apply(this, args);
				array._chartjs.listeners.forEach((object) => {
					if (typeof object[method] === 'function') {
						object[method](...args);
					}
				});
				return res;
			}
		});
	});
}
function unlistenArrayEvents(array, listener) {
	const stub = array._chartjs;
	if (!stub) {
		return;
	}
	const listeners = stub.listeners;
	const index = listeners.indexOf(listener);
	if (index !== -1) {
		listeners.splice(index, 1);
	}
	if (listeners.length > 0) {
		return;
	}
	arrayEvents.forEach((key) => {
		delete array[key];
	});
	delete array._chartjs;
}

function interpolateNumber(from, to, factor) {
  if (from === to) {
    return to;
  }
  return from + (to - from) * factor;
}

function interpolatorPoint(fromArray, i, to, factor) {
  const from = fromArray[i] || fromArray[i - 1] || fromArray._source;
  if (!from) {
    return to;
  }
  const x = interpolateNumber(from.x, to.x, factor);
  const y = interpolateNumber(from.y, to.y, factor);
  const angle = Number.isNaN(from.angle) ? interpolateNumber(from.angle, to.angle, factor) : undefined;
  return { x, y, angle };
}

function interpolatePoints(from, to, factor) {
  if (Array.isArray(from) && Array.isArray(to) && to.length > 0) {
    return to.map((t, i) => interpolatorPoint(from, i, t, factor));
  }
  return to;
}

function patchController(config, controller, elements = [], scales = []) {
  registry.addControllers(controller);
  registry.addElements(elements);
  registry.addScales(scales);
  config.type = controller.id;
  return config;
}

class GraphController extends ScatterController {
  constructor(chart, datasetIndex) {
    super(chart, datasetIndex);

    this._cachedEdgeOpts = {};
    this._edgeListener = {
      _onDataPush: (...args) => {
        const count = args.length;
        const start = this.getDataset().edges.length - count;
        const parsed = this._cachedMeta._parsedEdges;
        args.forEach((edge) => {
          parsed.push(this._parseDefinedEdge(edge));
        });
        this._insertEdgeElements(start, count);
      },
      _onDataPop: () => {
        this._cachedMeta.edges.pop();
        this._cachedMeta._parsedEdges.pop();
        this._scheduleResyncLayout();
      },
      _onDataShift: () => {
        this._cachedMeta.edges.shift();
        this._cachedMeta._parsedEdges.shift();
        this._scheduleResyncLayout();
      },
      _onDataSplice: (start, count, ...args) => {
        this._cachedMeta.edges.splice(start, count);
        this._cachedMeta._parsedEdges.splice(start, count);
        if (args.length > 0) {
          const parsed = this._cachedMeta._parsedEdges;
          parsed.splice(start, 0, ...args.map((edge) => this._parseDefinedEdge(edge)));
          this._insertEdgeElements(start, args.length);
        } else {
          this._scheduleResyncLayout();
        }
      },
      _onDataUnshift: (...args) => {
        const parsed = this._cachedMeta._parsedEdges;
        parsed.unshift(...args.map((edge) => this._parseDefinedEdge(edge)));
        this._insertEdgeElements(0, args.length);
      },
    };

    // this.resyncLayout();
  }

  initialize() {
    const type = this._type;
    const defaultConfig = defaults.get(type);
    this.edgeElementOptions = defaultConfig.edgeElementOptions;
    this.edgeElementType = registry.getElement(defaultConfig.edgeElementType);
    super.initialize();
    this._scheduleResyncLayout();
  }

  parse(start, count) {
    const meta = this._cachedMeta;
    const data = this._data;
    const iScale = meta.iScale;
    const vScale = meta.vScale;
    for (let i = 0; i < count; i++) {
      const index = i + start;
      const d = data[index];
      const v = meta._parsed[index] || {};
      if (d && typeof d.x === 'number') {
        v.x = d.x;
      }
      if (d && typeof d.y === 'number') {
        v.y = d.y;
      }
      meta._parsed[index] = v;
    }
    if (meta._parsed.length > data.length) {
      meta._parsed.splice(data.length, meta._parsed.length - data.length);
    }
    this._cachedMeta._sorted = false;
    iScale.invalidateCaches();
    vScale.invalidateCaches();

    this._parseEdges();
  }

  reset() {
    this.resetLayout();
    super.reset();
  }

  update(mode) {
    super.update(mode);

    const meta = this._cachedMeta;
    const edges = meta.edges || [];

    this.updateEdgeElements(edges, 0, mode);
  }

  destroy() {
    super.destroy();
    if (this._edges) {
      unlistenArrayEvents(this._edges, this._edgeListener);
    }
    this.stopLayout();
  }

  updateEdgeElements(edges, start, mode) {
    const bak = {
      _cachedDataOpts: this._cachedDataOpts,
      dataElementType: this.dataElementType,
      dataElementOptions: this.dataElementOptions,
      _sharedOptions: this._sharedOptions,
    };
    this._cachedDataOpts = {};
    this.dataElementType = this.edgeElementType;
    this.dataElementOptions = this.edgeElementOptions;
    const meta = this._cachedMeta;
    const nodes = meta.data;
    const data = meta._parsedEdges;

    const reset = mode === 'reset';

    const firstOpts = this.resolveDataElementOptions(start, mode);
    const sharedOptions = this.getSharedOptions(mode || 'normal', edges[start], firstOpts);
    const includeOptions = this.includeOptions(mode, sharedOptions);

    const xScale = meta.xScale;
    const yScale = meta.yScale;

    const base = {
      x: xScale.getBasePixel(),
      y: yScale.getBasePixel(),
    };

    function copyPoint(point) {
      const x = reset ? base.x : xScale.getPixelForValue(point.x);
      const y = reset ? base.y : yScale.getPixelForValue(point.y);
      return {
        x,
        y,
        angle: point.angle,
      };
    }

    for (let i = 0; i < edges.length; i++) {
      const edge = edges[i];
      const index = start + i;
      const parsed = data[index];

      const properties = {
        source: nodes[parsed.source],
        target: nodes[parsed.target],
        points: Array.isArray(parsed.points) ? parsed.points.map(copyPoint) : [],
      };
      properties.points._source = nodes[parsed.source];
      if (includeOptions) {
        properties.options = this.resolveDataElementOptions(index, mode);
      }
      this.updateEdgeElement(edge, index, properties, mode);
    }
    this.updateSharedOptions(sharedOptions, mode);

    Object.assign(this, bak);
  }

  updateEdgeElement(edge, index, properties, mode) {
    super.updateElement(edge, index, properties, mode);
  }

  updateElement(point, index, properties, mode) {
    if (mode === 'reset') {
      // start in center also in x
      const xScale = this._cachedMeta.xScale;
      properties.x = xScale.getBasePixel();
    }
    super.updateElement(point, index, properties, mode);
  }

  resolveNodeIndex(nodes, ref) {
    if (typeof ref === 'number') {
      // index
      return ref;
    }
    if (typeof ref === 'string') {
      // label
      const labels = this.chart.data.labels;
      return labels.indexOf(ref);
    }
    const nIndex = nodes.indexOf(ref);
    if (nIndex >= 0) {
      // hit
      return nIndex;
    }

    if (ref && typeof ref.index === 'number') {
      return ref.index;
    }

    const data = this.getDataset().data;
    const index = data.indexOf(ref);
    if (index >= 0) {
      return index;
    }

    console.warn('cannot resolve edge ref', ref);
    return -1;
  }

  buildOrUpdateElements() {
    const dataset = this.getDataset();
    const edges = dataset.edges || [];

    // In order to correctly handle data addition/deletion animation (an thus simulate
    // real-time charts), we need to monitor these data modifications and synchronize
    // the internal meta data accordingly.
    if (this._edges !== edges) {
      if (this._edges) {
        // This case happens when the user replaced the data array instance.
        unlistenArrayEvents(this._edges, this._edgeListener);
      }

      if (edges && Object.isExtensible(edges)) {
        listenArrayEvents(edges, this._edgeListener);
      }
      this._edges = edges;
    }
    super.buildOrUpdateElements();
  }

  draw() {
    const meta = this._cachedMeta;
    const edges = meta.edges || [];
    const elements = meta.data || [];

    const area = this.chart.chartArea;
    const ctx = this._ctx;

    if (edges.length > 0) {
      clipArea(ctx, area);
      edges.forEach((edge) => edge.draw(ctx, area));
      unclipArea(ctx);
    }

    elements.forEach((elem) => elem.draw(ctx, area));
  }

  _resyncElements() {
    super._resyncElements();

    const meta = this._cachedMeta;
    const edges = meta._parsedEdges;
    const metaEdges = meta.edges || (meta.edges = []);
    const numMeta = metaEdges.length;
    const numData = edges.length;

    if (numData < numMeta) {
      metaEdges.splice(numData, numMeta - numData);
      this._scheduleResyncLayout();
    } else if (numData > numMeta) {
      this._insertEdgeElements(numMeta, numData - numMeta);
    }
  }

  getTreeRootIndex() {
    const ds = this.getDataset();
    const nodes = ds.data;
    if (ds.derivedEdges) {
      // find the one with no parent
      return nodes.findIndex((d) => d.parent == null);
    }
    // find the one with no edge
    const edges = this._cachedMeta._parsedEdges || [];
    const nodeIndices = new Set(nodes.map((_, i) => i));
    edges.forEach((edge) => {
      nodeIndices.delete(edge.targetIndex);
    });
    return Array.from(nodeIndices)[0];
  }

  getTreeRoot() {
    const index = this.getTreeRootIndex();
    const p = this.getParsed(index);
    p.index = index;
    return p;
  }

  getTreeChildren(node) {
    const edges = this._cachedMeta._parsedEdges;
    return edges
      .filter((d) => d.source === node.index)
      .map((d) => {
        const p = this.getParsed(d.target);
        p.index = d.target;
        return p;
      });
  }

  _parseDefinedEdge(edge) {
    const ds = this.getDataset();
    const data = ds.data;
    return {
      source: this.resolveNodeIndex(data, edge.source),
      target: this.resolveNodeIndex(data, edge.target),
      points: [],
    };
  }

  _parseEdges() {
    const ds = this.getDataset();
    const data = ds.data;
    const meta = this._cachedMeta;
    if (ds.edges) {
      return (meta._parsedEdges = ds.edges.map((edge) => this._parseDefinedEdge(edge)));
    }

    const edges = (meta._parsedEdges = []);
    // try to derive edges via parent links
    data.forEach((node, i) => {
      if (node.parent != null) {
        // tree edge
        const parent = this.resolveNodeIndex(data, node.parent);
        edges.push({
          source: parent,
          target: i,
          points: [],
        });
      }
    });
    return edges;
  }

  addElements() {
    super.addElements();

    const meta = this._cachedMeta;
    const edges = this._parseEdges();
    const metaData = (meta.edges = new Array(edges.length));

    for (let i = 0; i < edges.length; ++i) {
      metaData[i] = new this.edgeElementType();
    }
  }

  _resyncEdgeElements() {
    const meta = this._cachedMeta;
    const edges = this._parseEdges();
    const metaData = meta.edges || (meta.edges = []);

    for (let i = 0; i < edges.length; ++i) {
      metaData[i] = metaData[i] || new this.edgeElementType();
    }
    if (edges.length < metaData.length) {
      metaData.splice(edges.length, metaData.length);
    }
  }

  _insertElements(start, count) {
    super._insertElements(start, count);
    if (count > 0) {
      this._resyncEdgeElements();
    }
  }

  _removeElements(start, count) {
    super._removeElements(start, count);
    if (count > 0) {
      this._resyncEdgeElements();
    }
  }

  _insertEdgeElements(start, count) {
    const elements = [];
    for (let i = 0; i < count; i++) {
      elements.push(new this.edgeElementType());
    }
    this._cachedMeta.edges.splice(start, 0, ...elements);
    this.updateEdgeElements(elements, start, 'reset');
    this._scheduleResyncLayout();
  }

  _onDataPush() {
    super._onDataPush.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }
  _onDataPop() {
    super._onDataPop();
    this._scheduleResyncLayout();
  }
  _onDataShift() {
    super._onDataShift();
    this._scheduleResyncLayout();
  }
  _onDataSplice() {
    super._onDataSplice.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }
  _onDataUnshift() {
    super._onDataUnshift.apply(this, Array.from(arguments));
    this._scheduleResyncLayout();
  }

  reLayout() {
    // hook
  }

  resetLayout() {
    // hook
  }

  stopLayout() {
    // hook
  }

  _scheduleResyncLayout() {
    if (this._scheduleResyncLayoutId != null) {
      return;
    }
    this._scheduleResyncLayoutId = requestAnimFrame.call(window, () => {
      this._scheduleResyncLayoutId = null;
      this.resyncLayout();
    });
  }

  resyncLayout() {
    // hook
  }
}

GraphController.id = 'graph';
GraphController.defaults = /*#__PURE__*/ merge({}, [
  ScatterController.defaults,
  {
    datasets: {
      clip: 10, // some space in combination with padding
      animation: {
        points: {
          fn: interpolatePoints,
          properties: ['points'],
        },
      },
    },
    layout: {
      padding: 10,
    },
    scales: {
      x: {
        display: false,
        ticks: {
          maxTicksLimit: 2,
          precision: 100,
          minRotation: 0,
          maxRotation: 0,
        },
      },
      y: {
        display: false,
        ticks: {
          maxTicksLimit: 2,
          precision: 100,
          minRotation: 0,
          maxRotation: 0,
        },
      },
    },
    tooltips: {
      callbacks: {
        label(item) {
          return item.chart.data.labels[item.dataIndex];
        },
      },
    },
    edgeElementType: EdgeLine.id,
    edgeElementOptions: Object.assign(
      {
        tension: 'lineTension',
        stepped: 'lineStepped',
        directed: 'directed',
        arrowHeadSize: 'arrowHeadSize',
        arrowHeadOffset: 'pointRadius',
      },
      (() => {
        const options = {};
        LineController.defaults.datasetElementOptions.forEach((attr) => {
          options[attr] = `line${attr[0].toUpperCase()}${attr.slice(1)}`;
        });
        return options;
      })()
    ),
  },
]);

class GraphChart extends Chart {
  constructor(item, config) {
    super(item, patchController(config, GraphController, [EdgeLine, Point], LinearScale));
  }
}
GraphChart.id = GraphController.id;

class ForceDirectedGraphController extends GraphController {
  constructor(chart, datasetIndex) {
    super(chart, datasetIndex);
    this._simulation = forceSimulation()
      .on('tick', () => {
        this._copyPosition();
        this.chart.render();
      })
      .on('end', () => {
        this._copyPosition();
        this.chart.render();
      });
    const sim = this._config.simulation;

    const fs = {
      center: forceCenter,
      collide: forceCollide,
      link: forceLink,
      manyBody: forceManyBody,
      x: forceX,
      y: forceY,
      radial: forceRadial,
    };

    Object.keys(fs).forEach((key) => {
      const options = sim.forces[key];
      if (!options) {
        return;
      }
      const f = fs[key]();
      if (typeof options !== 'boolean') {
        Object.keys(options).forEach((attr) => {
          f[attr](options[attr]);
        });
      }
      this._simulation.force(key, f);
    });
    this._simulation.stop();
  }

  _copyPosition() {
    const nodes = this._cachedMeta._parsed;

    const minmax = nodes.reduce(
      (acc, v) => {
        const s = v._sim;
        if (!s) {
          return acc;
        }
        if (s.x < acc.minX) {
          acc.minX = s.x;
        }
        if (s.x > acc.maxX) {
          acc.maxX = s.x;
        }
        if (s.y < acc.minY) {
          acc.minY = s.y;
        }
        if (s.y > acc.maxY) {
          acc.maxY = s.y;
        }
        return acc;
      },
      {
        minX: Number.POSITIVE_INFINITY,
        maxX: Number.NEGATIVE_INFINITY,
        minY: Number.POSITIVE_INFINITY,
        maxY: Number.NEGATIVE_INFINITY,
      }
    );

    const rescaleX = (v) => ((v - minmax.minX) / (minmax.maxX - minmax.minX)) * 2 - 1;
    const rescaleY = (v) => ((v - minmax.minY) / (minmax.maxY - minmax.minY)) * 2 - 1;

    nodes.forEach((node) => {
      if (node._sim) {
        node.x = rescaleX(node._sim.x);
        node.y = rescaleY(node._sim.y);
      }
    });

    const xScale = this._cachedMeta.xScale;
    const yScale = this._cachedMeta.yScale;
    const elems = this._cachedMeta.data;
    elems.forEach((elem, i) => {
      const parsed = nodes[i];
      elem.x = xScale.getPixelForValue(parsed.x, i);
      elem.y = yScale.getPixelForValue(parsed.y, i);
    });
  }

  resetLayout() {
    super.resetLayout();
    this._simulation.stop();

    const nodes = this._cachedMeta._parsed.map((node, i) => {
      const simNode = Object.assign({}, node);
      simNode.index = i;
      node._sim = simNode;
      if (!node.reset) {
        return simNode;
      }
      delete simNode.x;
      delete simNode.y;
      delete simNode.vx;
      delete simNoe.vy;
      return simNode;
    });
    this._simulation.nodes(nodes);
    this._simulation.alpha(1).restart();
  }

  resyncLayout() {
    super.resyncLayout();
    this._simulation.stop();

    const meta = this._cachedMeta;

    const nodes = meta._parsed.map((node, i) => {
      const simNode = Object.assign({}, node);
      simNode.index = i;
      node._sim = simNode;
      if (simNode.x === null) {
        delete simNode.x;
      }
      if (simNode.y === null) {
        delete simNode.y;
      }
      if (simNode.x == null && simNode.y == null) {
        node.reset = true;
      }
      return simNode;
    });
    const link = this._simulation.force('link');
    if (link) {
      link.links([]);
    }
    this._simulation.nodes(nodes);
    if (link) {
      // console.assert(ds.edges.length === meta.edges.length);
      // work on copy to avoid change
      link.links((meta._parsedEdges || []).map((link) => Object.assign({}, link)));
    }

    if (this._config.simulation.initialIterations > 0) {
      this._simulation.alpha(1);
      this._simulation.tick(this._config.simulation.initialIterations);
      this._copyPosition();
      if (this._config.simulation.autoRestart) {
        this._simulation.restart();
      } else {
        requestAnimFrame.call(window, () => this.chart.update());
      }
    } else if (this._config.simulation.autoRestart) {
      this._simulation.alpha(1).restart();
    }
  }

  reLayout() {
    this._simulation.alpha(1).restart();
  }

  stopLayout() {
    super.stopLayout();
    this._simulation.stop();
  }
}

ForceDirectedGraphController.id = 'forceDirectedGraph';
ForceDirectedGraphController.defaults = /*#__PURE__*/ merge({}, [
  GraphController.defaults,
  {
    animation: false,
    datasets: {
      simulation: {
        initialIterations: 0,
        autoRestart: true,
        forces: {
          center: true,
          collide: false,
          link: true,
          manyBody: true,
          x: false,
          y: false,
          radial: false,
        },
      },
    },
    scales: {
      x: {
        min: -1,
        max: 1,
      },
      y: {
        min: -1,
        max: 1,
      },
    },
  },
]);

class ForceDirectedGraphChart extends Chart {
  constructor(item, config) {
    super(item, patchController(config, ForceDirectedGraphController, [EdgeLine, Point], LinearScale));
  }
}
ForceDirectedGraphChart.id = ForceDirectedGraphController.id;

class DendogramController extends GraphController {
  updateEdgeElement(line, index, properties, mode) {
    properties._orientation = this._config.tree.orientation;
    super.updateEdgeElement(line, index, properties, mode);
  }

  updateElement(point, index, properties, mode) {
    if (index != null) {
      properties.angle = this.getParsed(index).angle;
    }
    super.updateElement(point, index, properties, mode);
  }

  resyncLayout() {
    const meta = this._cachedMeta;

    meta.root = hierarchy(this.getTreeRoot(), (d) => this.getTreeChildren(d))
      .count()
      .sort((a, b) => b.height - a.height || b.data.index - a.data.index);

    this.doLayout(meta.root);

    super.resyncLayout();
  }

  reLayout(newOptions) {
    if (newOptions) {
      Object.assign(this._config.tree, newOptions);
      const ds = this.getDataset();
      if (ds.tree) {
        Object.assign(ds.tree, newOptions);
      } else {
        ds.tree = newOptions;
      }
    }
    this.doLayout(this._cachedMeta.root);
  }

  doLayout(root) {
    const options = this._config.tree;

    const layout = options.mode === 'tree' ? tree() : cluster();

    if (options.orientation === 'radial') {
      layout.size([Math.PI * 2, 1]);
    } else {
      layout.size([2, 2]);
    }

    const orientation = {
      horizontal: (d) => {
        d.data.x = d.y - 1;
        d.data.y = -d.x + 1;
      },
      vertical: (d) => {
        d.data.x = d.x - 1;
        d.data.y = -d.y + 1;
      },
      radial: (d) => {
        d.data.x = Math.cos(d.x) * d.y;
        d.data.y = Math.sin(d.x) * d.y;
        d.data.angle = d.y === 0 ? Number.NaN : d.x;
      },
    };

    layout(root).each(orientation[options.orientation] || orientation.horizontal);

    requestAnimFrame.call(window, () => this.chart.update());
  }
}

DendogramController.id = 'dendogram';
DendogramController.defaults = /*#__PURE__*/ merge({}, [
  GraphController.defaults,
  {
    datasets: {
      tree: {
        mode: 'dendogram', // dendogram, tree
        orientation: 'horizontal', // vertical, horizontal, radial
      },
      animation: {
        numbers: {
          type: 'number',
          properties: ['x', 'y', 'angle', 'radius', 'rotation', 'borderWidth'],
        },
      },
      tension: 0.4,
    },
    scales: {
      x: {
        min: -1,
        max: 1,
      },
      y: {
        min: -1,
        max: 1,
      },
    },
  },
]);

class DendogramChart extends Chart {
  constructor(item, config) {
    super(item, patchController(config, DendogramController, EdgeLine));
  }
}
DendogramChart.id = DendogramController.id;

class TreeController extends DendogramController {}

TreeController.id = 'tree';
TreeController.defaults = /*#__PURE__*/ merge({}, [
  DendogramController.defaults,
  {
    datasets: {
      tree: {
        mode: 'tree',
      },
    },
  },
]);

class TreeChart extends Chart {
  constructor(item, config) {
    super(item, patchController(config, TreeController, [EdgeLine, Point], LinearScale));
  }
}
TreeChart.id = TreeController.id;

export { DendogramChart, DendogramController, EdgeLine, ForceDirectedGraphChart, ForceDirectedGraphController, GraphChart, GraphController, TreeChart, TreeController };
